
{
 By    : TOTH Paul tothpaul@mygale.org
 Page  : http://www.mygale.org/~tothpaul
 From  : Martin Austermeier (?)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 Bugs :
  - It doesn't work with Word 7.0 under Win95 (32 bit).
  - when FileNew is used on a unexisting file, Word display a message that need a KeyPress.

 Any help :
  - Does someone know how to talk to Word7.0
  - Does someone know how to handle Word Error messages.

}

unit Wordcmd;

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs;

type
  TArrayDef = RECORD
    cArrayDimensions : Integer;
    arrayDimensions : Array[0..0] OF Byte;
  END;
  PArrayDef = ^TArrayDef;

  PDoubleArray = Pointer;
  PStringArray = ^PChar;

CONST
  MAX_ARGS = 34;  { MaxArgs based on largest dialog }

Type
  TParmKind = (pkNone, pkShort, pkLong, pkDouble, pkString, pkWrite, pkRead );
  TParmFlag = Set of TParmKind;

  TParm = RECORD  { WDOPR }
    Value : RECORD
    CASE Integer OF
      0 : (AsShort  : Integer);
      1 : (AsLong   : LongInt);
      2 : (AsDouble : Double);
      3 : (AsString : PChar);
      4 : (AsArray  : PArrayDef;
           Ptr : RECORD
           CASE Boolean OF
             FALSE : (DoubleArray : PDoubleArray);
             TRUE  : (StringArray : PStringArray);
           END;
          );
    END;
    Size  : Word;
    Data  : record
     Case Byte of
      0 : (Kind  : TParmKind);
      1 : (Flag  : TParmFlag);
      2 : (Value : Integer); { so, Field is 2 bytes long }
    end;
    FldID : Word;
   END;

  TParmEditor = class
   Private
    Parm:^TParm;
    Procedure SetDataKind(Value:TParmKind);
    Procedure SetAsShort (Value:Integer);
    Procedure SetAsLong  (Value:LongInt);
    Procedure SetAsDouble(Value:Double);
    Procedure SetAsString(Value:PChar);
   Public
    Property AsShort:Integer Write SetAsShort;
    Property AsLong:LongInt  Write SetAsLong;
    Property AsDouble:double Write SetAsDouble;
    Property AsString:PChar  Write SetAsString;
    Procedure SetFieldID(Value:integer);
    Procedure GetFieldID(Value:integer);
  end;

  TParmList = Array[0..MAX_ARGS-1] of TParm;

  TWordMsg = RECORD
   CmdID     : Integer;
   DlgOpts   : Integer;
   ParmCount : Integer;
   Parms     : ^TParmList;
   Return    : ^TParm;
  END;


  TShowCmd = (swHide, swMinimize, swRestore, swShow, swShowMaximized, swShowMinimized,
              swShowMinNoActive, swShowNA, swShowNoActivate, swShowNormal);

  TWordErrorEvent  = Procedure ( Sender : TObject; Error : Integer) of Object;
  EWordError = class (Exception);

  TWordCmd = class(TComponent)
  private
    { Private-dclarations }
    WordMsg    : TWordMsg;
    fReturn    : ^TParm;
    fActive    : boolean;
    fParmCount : Word;
    fParmSize  : Word;
    ParmEditor : TParmEditor;
    fWordHandle:HWnd;
    fShell:boolean;
    fShellCommand:string;
    fShowCmd:TShowCmd;
    EOnError:TWordErrorEvent;
    Function GetActive:boolean;
    Procedure SetActive(Value:boolean);
    Procedure SetShell(Value:boolean);
    Procedure RegisteredShell;
    Procedure ShellWord;
    Function GetParm(Index:Integer):TParmEditor;
    Function GetReturn:TParmEditor;
    Procedure SetParmCount(Value:Word);
    Procedure SetParmSize(Value:Word);
    Procedure SendWordMessage(MsgId:Integer);
    Procedure SendWordDlgMessage(MsgId:Integer);
    Procedure SetReply(Value:Integer);
  protected
    { Protected-dclarations }
    Constructor Create(AOwner:TComponent); override;
    Destructor Destroy; override;
    Procedure Loaded; override;
    Property ParmCount:Word read fParmCount Write SetParmCount;
    Property ParmSize :Word read fParmSize  Write SetParmSize;
    Property Reply:Integer Write SetReply;
    Property Parms[Index:Integer]:TParmEditor read GetParm;
    Property Return:TParmEditor read GetReturn;
  public
    { Public-dclarations }
    Function  ErrorMessage(Error:Integer):string;
    Procedure OpenWord;
    Procedure FileNew;
    Procedure FileOpen(Name:string);
    Procedure FileSave;
    Procedure FileSaveAs(Name:String);
    Procedure FileClose(Save:boolean);
    Procedure FilePrint;
    Procedure EditPaste;
    Function  EditFindFound:boolean;
    Function  EditSearch(Text:String):boolean;
    Function  EditReplace(Search,Replace:String; All:Boolean):boolean;
    Procedure Insert(Text:String);
    Procedure StartOfDoc;
    Procedure CloseWord;
  published
    { Published-dclarations }
    Property Active:boolean read GetActive write SetActive;
    Property Shell:boolean read fShell write SetShell;
    Property ShellCommand:string read fShellCommand write fShellCommand stored false;
    Property ShowCmd:TShowCmd read fShowCmd write fShowCmd default swMinimize;
    Property OnError:TWordErrorEvent read EOnError write EOnError;
  end;

procedure Register;

implementation

Uses
 ShellAPI,DsgnIntf, WdCmds, WdFid;

{ TParmEditor }

Procedure TParmEditor.SetDataKind;
 begin
  Parm^.Data.Value:=Ord(Value);
  Parm^.Size:=0;
  Parm^.FldID:=0;
 end;
Procedure TParmEditor.SetAsShort;
 begin
  SetDataKind(pkShort);
  Parm^.Value.AsShort:=Value;
 end;
Procedure TParmEditor.SetAsLong;
 begin
  SetDataKind(pkLong);
  Parm^.Value.AsLong:=Value;
 end;
Procedure TParmEditor.SetAsDouble;
 begin
  SetDataKind(pkDouble);
  Parm^.Value.AsDouble:=Value;
 end;
Procedure TParmEditor.SetAsString;
 begin
  SetDataKind(pkString);
  Parm^.Value.AsString:=Value;
  Parm^.Size:=StrLen(Value);
 end;
Procedure TParmEditor.SetFieldID;
 begin
  Include(Parm^.Data.Flag,pkWrite);
  Parm^.FldID:=Value;
 end;
Procedure TParmEditor.GetFieldID;
 begin
  Include(Parm^.Data.Flag,pkRead);
  Parm^.FldID:=Value;
 end;

{ TWordCmd }

Const
 WM_WORDMESSAGE = WM_USER + $0300;

Constructor TWordCmd.Create(AOwner:TComponent);
 begin
  inherited Create(AOwner);
  fActive    :=False;
  fWordHandle:=0;
  fShowCmd   :=swMinimize;
  fParmCount :=0;
  fParmSize  :=0;
  fShell     :=False;
  Active     :=True;
  fReturn    :=nil;
  ParmEditor :=TParmEditor.Create;
 end;

Destructor TWordCmd.Destroy;
 begin
  if fReturn<>nil then Dispose(fReturn);
  ParmEditor.Free;
  ParmSize:=0;
  Inherited Destroy;
 end;

Procedure TWordCmd.Loaded;
 begin
  Inherited Loaded;
  if fActive then
   Active:=True
  else
   fWordHandle:=FindWindow('OpusApp',nil);
 end;

Function TWordCmd.GetActive;
 begin
  Result:=(fWordHandle<>0)and(fWordHandle=FindWindow('OpusApp',nil));
 end;

Procedure TWordCmd.SetActive;
 begin
  fActive:=Value;
  if Value=False then
   fWordHandle:=0
  else begin
   fWordHandle:=FindWindow('OpusApp',nil);
   if (fWordHandle=0)and(fShell) then ShellWord;
  end;
 end;

Procedure TWordCmd.SetShell;
 begin
  if fShell=Value then exit;
  fShell:=Value;
  if (fShell)and(fShellCommand='') then RegisteredShell;
 end;

Procedure TWordCmd.RegisteredShell;
 Var
  Size:LongInt;
  ShellName:Array[0..79] of char;
  ShellOpen:Array[0..79] of char;
  Space:PChar;
 begin
  Size:=79;
  if RegQueryValue( HKEY_CLASSES_ROOT, '.doc', ShellName, Size)=0 then begin
   StrCat(ShellName,'\Shell\Open\Command');
   Size:=79;
   If RegQueryValue( HKEY_CLASSES_ROOT, ShellName, ShellOpen, Size)=0 then
    Space:=StrPos(ShellOpen,' '); if Space<>nil then Space^:=#0;
    ShellCommand:=StrPas(ShellOpen);
  end;
 end;

Procedure TWordCmd.ShellWord;
 Const
  SWShowCmd : Array[TShowCmd] of integer =
  (sw_Hide, sw_Minimize, sw_Restore, sw_Show, sw_ShowMaximized, sw_ShowMinimized,
   sw_ShowMinNoActive, sw_ShowNA, sw_ShowNoActivate, sw_ShowNormal);
 Var
  OpenCmd:Array[0..79] of char;
  Shell:integer;
 begin
  if fShellCommand<>'' then
   StrPCopy(OpenCmd,fShellCommand)
  else
   StrCopy(OpenCmd,'WINWORD.EXE');
  Shell:=ShellExecute(0,nil,OpenCmd,nil,nil,SWShowCmd[fShowCmd]);
  If Shell<32 then ShowMessage('Word non lanc #'+IntToHex(Shell,4));
  fWordHandle:=FindWindow('OpusApp',nil);
 end;

Function TWordCmd.GetParm(Index:Integer):TParmEditor;
 begin
  If Index>=ParmCount then ParmCount:=Index+1;
  ParmEditor.Parm:=@WordMsg.Parms^[Index];
  Result:=ParmEditor;
 end;

Function TWordcmd.GetReturn:TParmEditor;
 begin
  if fReturn=nil then New(fReturn);
  WordMsg.Return:=@fReturn^;
  ParmEditor.Parm:=@fReturn^;
  Result:=ParmEditor;
 end;

Procedure TWordCmd.SetParmCount(Value:Word);
 begin
  WordMsg.ParmCount:=Value;
  WordMsg.Return:=nil;
  If fParmCount<Value then begin
   ParmSize:=Value*SizeOf(TParm);
   fParmCount:=Value;
  end;
 end;

Procedure TWordCmd.SetParmSize(Value:Word);
 Var
  NewParms:pointer;
 begin
  If Value>0 then GetMem(NewParms,Value) else NewParms:=nil;
  if fParmSize>0 then begin
   if Value>0 then Move(WordMsg.Parms^,NewParms^,Value);
   FreeMem(WordMsg.Parms,fParmSize);
  end;
  fParmSize:=Value;
  WordMsg.Parms:=NewParms;
 end;

Procedure TWordCmd.SendWordMessage(MsgId:integer);
 begin
  WordMsg.CmdId   := MsgId;
  WordMsg.DlgOpts := 0;
  If Active then
   Reply:=SendMessage(fWordHandle,WM_WORDMESSAGE,0,LongInt(@WordMsg))
  else
   ShowMessage('Winword not found');
 end;

Procedure TWordCmd.SendWordDlgMessage(MsgId:integer);
 Type
  ADlgOption = (CMD_DEFAULTS,     { GetCurValues }
                CMD_DIALOG,       { display dialog }
                CMD_ACTION,       { execute dialog }
                CMD_DLG_ACTION);  { display & exec }
 begin
  WordMsg.CmdId   := MsgId;
  WordMsg.DlgOpts := Integer(CMD_ACTION);
  If Active then
   Reply:=SendMessage(fWordHandle,WM_WORDMESSAGE,0,LongInt(@WordMsg))
  else
   ShowMessage('Winword not found');
 end;

Procedure TWordCmd.SetReply;
 begin
 {  -1 = FileExit }
  if Value>0 then
   If Assigned(EOnError) then
    EOnError(Self,Value)
   else
    raise EWordError.Create(ErrorMessage(Value));
 end;

{>> Word Dialog <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<}

Function PasToStr(Var s:string):PChar;
 begin
  s[Length(s)+1]:=#0;
  Result:=@s[1];
 end;

Function TWordCmd.ErrorMessage(Error:Integer):string;
 begin
  Case Error of
    102 : Result:='User Cancel ?';
   1078 : Result:='File not found';
   5003 : Result:='Unabled to execute command (no file open ?)';
   5004 : Result:='Invalid parameters count';
   5007 : Result:='No result for a procedure';
   5008 : Result:='Invalid parameter';
   5016 : Result:='Field not found';
   5020 : Result:='Integer parameter expected';
   5023 : Result:='Result is a number';
   5024 : Result:='Result is a string';
   5030 : Result:='Invalid pointer in message';
   else Result:='Unknow error #'+IntToStr(Error);
  end;
 end;

{
 Two kind of messages hare allowed :
  - SendWordMessage : standard calls
  - SendWordDlgMessage : for dialog messages
 I realy don't know when to use one or the other...just try the other one when
 the first wan't past.

 To set parameters :
  - Set ParmCount (memory is also allocated)
  - Set Parms[x].AsYYYY := ZZZZ;
  - for Dialogs, set FieldId :
   . SetFieldID(IIIIII) to set a field
   . GetFieldID(IIIIII) to read a field

 After a SendWordDlgMEssage, declared Fields are set by Word
}

Procedure TWordCmd.OpenWord;
 begin
  Active:=True; { No message send ! }
 end;

Procedure TWordCmd.FileNew;
 begin
  ParmCount:=0;
  SendWordDlgMessage(wdFileNew);
 end;

Procedure TWordCmd.FileOpen;
 begin
  ParmCount:=1;
  Parms[0].AsString:=PasToStr(Name);
  SendWordDlgMessage(wdFileOpen);
 end;

Procedure TWordCmd.FileSave;
 begin
  ParmCount:=0;
  SendWordMessage(wdFileSave);
 end;

Procedure TWordCmd.FileSaveAs(Name:string);
 begin
  ParmCount:=1;
  Parms[0].AsString:=PasToStr(Name);
  SendWordDlgMessage(wdFileSaveAs);
 end;

Procedure TWordCmd.FileClose;
 begin
  ParmCount:=1;
  Parms[0].AsShort:=2-(Integer(Save)); { 0=Dialog, 1=Oui, 2=Non }
  SendWordMessage(wdFileClose);
 end;

Procedure TWordCmd.FilePrint;
 begin
  ParmCount:=0;
  SendWordDlgMessage(wdFilePrint);
 end;

Procedure TWordCmd.EditPaste;
 begin
  ParmCount:=0;
  SendWordMessage(wdEditPaste);
 end;

Function TWordCmd.EditFindFound:boolean;
 begin
  ParmCount:=0;
  Return.AsShort:=0;
  SendWordMessage(wdEditFindFound);
  Result:=Boolean(-WordMsg.Return^.Value.AsShort);
 end;

Function TWordCmd.EditSearch(Text:String):boolean;
 begin
  ParmCount:=1;
  Parms[0].AsString:=PasToStr(Text);
  SendWordDlgMessage(wdEditFind);
  Result:=EditFindFound;
 end;

Function TWordCmd.EditReplace(Search,Replace:string; All:boolean):boolean;
 begin
  ParmCount:=4;
  Parms[0].AsString:=PasToStr(Search);
  Parms[1].AsString:=PasToStr(Replace);
  Parms[2].AsShort:=-1;
  Parms[2].SetFieldId(fidReplaceOne);
  Parms[3].AsShort:=-(Integer(All)); { Basic(True) = -1; Pascal(True)=+1 }
  Parms[3].SetFieldId(fidReplaceAll);
  SendWordDlgMessage(wdEditReplace);
  Result:=All or EditFindFound;
 end;

Procedure TWordCmd.Insert(Text:string);
 begin
  ParmCount:=1;
  Parms[0].AsString:=PasToStr(Text);
  SendWordMessage(wdInsert);
 end;

Procedure TWordCmd.StartOfDoc;
 begin
  ParmCount:=1;
  Parms[0].AsString:='\StartOfDoc';
  SendWordDlgMessage(wdEditGoto);
 end;

Procedure TWordCmd.CloseWord;
 begin
  ParmCount:=1;
  Parms[0].AsShort:=2; { no save }
  SendWordMessage(wdFileExit);
 end;

 { Registration }

procedure Register;
begin
  RegisterComponents('Additional', [TWordCmd]);
end;

end.
